# Cross-page constants to import from here to avoid circular imports

LEADERBOARD_PAGE_NAME = "Leaderboard"
RECORDS_PAGE_NAME = "Records"
COMPARE_PAGE_NAME = "Compare"

RECORDS_LIMIT = 1000

CACHE_TTL = 60 * 15  # 15 minutes

PINNED_COL_NAME = "trulens.dashboard.pinned"
EXTERNAL_APP_COL_NAME = "trulens.dashboard.external_app"
HIDE_RECORD_COL_NAME = "trulens.dashboard.hide_record"
